{-# OPTIONS -fglasgow-exts #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE OverlappingInstances #-}
{-# OPTIONS -fno-cse #-}

module Control.Parallel.Eden.Grace (
    Placeable, Placeable',
    Node(N), NodeList, 
    SimpleEdge, Edge(E), EdgeList,
    Function, FunctionList,
    toFL, number, numberFrom,
    toEL, fromEL, 
    nothings,
    Lister, lister,
    ProcessNetwork, build, start,
    checkTopology, isTopologyOK
)where

import Control.Monad
import Data.List (find, groupBy, sortBy, zipWith4, elemIndex)
import Data.Maybe (fromJust)
import Debug.Trace
import Control.Parallel.Eden
import Control.Parallel.Eden.Grace.IsFunctionType
import System.IO.Unsafe
import Unsafe.Coerce


---------------------------------------------------------------------

-- classes and instances for the placeable function

-- | The class 'Placeable' contains methods that are needed when placing a function on a node.
class (Trans argtype, 
       Trans restype) => 
      Placeable ftype argtype remtype restype plisttype
        | ftype -> argtype remtype restype plisttype where
        -- | Create a channel and the incoming result for the first parameter of the function
        createChan :: ftype -> (ChanName argtype, argtype)
        -- | Apply the first parameter to the function
        apply :: ftype -> argtype -> remtype
        -- | Cast a channel to the correct type for this function's (final) result
        castOutputChan :: ftype -> ChanName a -> ChanName restype
        -- | Determine whether the function is a constant or takes at least one parameter
        isConst :: ftype -> Bool
        -- | Apply all parameters to this function
        applyAll :: ftype -> plisttype -> restype
        -- | Generate channels and incoming results for all parameters of the function
        generateChannels :: ftype -> ([ChanName Dummy], plisttype)


-- | This instance of 'Placeable' delegates all function calls to an appropriate instance
-- of 'Placeable''
instance (IsFunctionType ftype flag, 
          Placeable' flag ftype argtype remtype restype plisttype) => 
         Placeable ftype argtype remtype restype plisttype where
        createChan = createChan' (undefined::flag)  --oder doch lieber Fehler schmeissen?
        apply = apply' (undefined::flag)
        castOutputChan = castOutputChan' (undefined::flag)
        isConst = isConst' (undefined::flag)
        applyAll = applyAll' (undefined::flag)
        generateChannels = generateChannels' (undefined::flag)

-- | The class 'Placeable'' mirrors the class 'Placeable'. Each function takes an additional 
-- parameter whose type (one of 'HConst', 'HFuncConstParam' or 'HListParam' ) gives extra 
-- information about the type of the second parameter.
class (Trans argtype, 
       Trans restype) => 
      Placeable' flag ftype argtype remtype restype plisttype 
        | flag ftype -> argtype remtype restype plisttype where
        createChan' :: flag -> ftype -> (ChanName argtype, argtype)
        apply' :: flag -> ftype -> argtype -> remtype
        castOutputChan' :: flag -> ftype -> ChanName a -> ChanName restype
        isConst' :: flag -> ftype -> Bool
        applyAll' :: flag -> ftype -> plisttype -> restype
        generateChannels' :: flag -> ftype -> ([ChanName Dummy], plisttype)
        generateListerChannels' :: flag -> ftype -> [Int] -> ([ChanName Dummy], plisttype)

-- | An instance for constant functions (that take no parameters).
instance (Trans ftype) => Placeable' HConst ftype () () ftype PNilType where
        createChan' = trace "error: called createChan' on constant" $ error "no chan creation for constant function"
        apply' = trace "error: called apply' for constant" $ error "no application of constant function"
        castOutputChan' _ _ = unsafeCoerce
        isConst' _ _ = True
        applyAll' _ f PNil = f
        generateChannels' _ _ = ([], PNil)
        generateListerChannels' _ _ [] = ([], PNil)
        generateListerChannels' _ _ [0] = ([], PNil)
        generateListerChannels' _ _ _ = error "can not generate channels for constant"

-- | An instance for ordinary functions (that take at least one parameter) where the first parameter
-- is not a list.
instance (Trans argtype,
          --Placeable remtype a g restype plisttype,
          Placeable' flag remtype a g restype plisttype,
          IsFunctionType remtype flag)
          => Placeable' HFuncConstParam (argtype -> remtype) argtype remtype restype (PConsType argtype plisttype) where
        createChan' _ _ = new (\name val -> (name, val))
        apply' _ f x = f x
        castOutputChan' _ _ = unsafeCoerce
        isConst' _ _ = False
        applyAll' myFlag f (PCons x xs) = applyAll' (undefined::flag) (apply' myFlag f x) xs
        generateChannels' myFlag f = generateListerChannels' myFlag f [0] 
        generateListerChannels' myFlag f [] = generateListerChannels' myFlag f [0]
        generateListerChannels' myFlag f (0:ns) = ((unsafeCoerce name):msgs, PCons val vals) 
                where
                    (name, val) = createChan' myFlag f
                    (msgs, vals) = generateListerChannels' (undefined:: flag) (undefined::remtype) ns

        generateListerChannels' _ _ (n:ns) = error $ "can not generate "++show n++" channels for constant param"

-- | An instance for ordinary functions (that take at least one parameter) where the first parameter
-- is a list.
instance (Trans argelemtype,
          --Placeable remtype a g restype plisttype,
          Placeable' flag remtype a g restype plisttype,
          IsFunctionType remtype flag)
          => Placeable' HFuncListParam ([argelemtype] -> remtype) [argelemtype] remtype restype (PConsType [argelemtype] plisttype) where
        createChan' _ _ = new (\name val -> (name, val))
        apply' _ f x = f x
        castOutputChan' _ _ = unsafeCoerce
        isConst' _ _ = False
        applyAll' myFlag f (PCons x xs) = applyAll' (undefined::flag) (apply' myFlag f x) xs
        generateChannels' myFlag f = generateListerChannels' myFlag f [0]
        generateListerChannels' myFlag f [] = generateListerChannels' myFlag f [0]
        generateListerChannels' myFlag f (0:ns) = ((unsafeCoerce name):msgs, PCons val vals) 
                where
                    (name, val) = createChan' myFlag f
                    (msgs, vals) = generateListerChannels' (undefined:: flag) (undefined::remtype) ns
        
        generateListerChannels' myFlag f (n:ns) = (allChans, PCons myVal otherVals)
            where
                (otherChans, otherVals) = generateListerChannels' (undefined::flag) (undefined::remtype) ns
                (allChans, myVal) = gen n otherChans []
            
                gen :: Int -> [ChanName Dummy] -> [argelemtype] -> ([ChanName Dummy], [argelemtype])
                gen 0 cs l = (cs, l)
                gen n cs l = new (\name val -> gen (n-1) (unsafeCoerce name:cs) (val:l))


-- | A type constructor for the parameter list
data PNilType = PNil
-- | A type constructor for the parameter list
data PConsType a b = PCons a b

-- | Wrapper for a placeable function to allow a heterogeneous list
data Function = forall f a g r p. (Placeable f a g r p) => F f
-- | A list of placeable functions.
type FunctionList = [Function] 

-- | A node with a label and a placed function
data Node l = forall f a g r p. (Placeable f a g r p) => N l f
-- | A list of nodes
type NodeList l = [Node l] 

type SimpleEdge n l = (n, n, l) -- (from, to, label)

-- | An edge that may carry a filter function
data Edge n l = forall a b p. (Trans a, Trans b, Placeable (a -> b) a b b p) 
              => E 
                 n                -- the node where the edge starts
                 n                -- the node where the edge ends
                 l                -- the label of the edge
                 (Maybe (a -> b)) -- the carried filter function or Nothing
-- | A list of edges
type EdgeList n l = [Edge n l]

-- | A Show instance for edges
instance (Show n, Show l) => Show (Edge n l) where
    show (E f t l _) = "(E "++show f++" "++show t++" "++show l++" _)"

---------------------------------------------------------------------
-- function list helpers

-- | Convert a (homogeneous) list of functions into a 'FunctionList'
toFL :: (Placeable f a g r p) => [f] -> FunctionList
toFL = map F 

-- | Creates a list of nodes with Int-labels beginning with 1 and places the given functions on them.
number :: FunctionList -> NodeList Int
number = numberFrom 1

-- | Creates a list of nodes with Int-labels beginning with the given parameter and places the given functions on them.
numberFrom :: Int -> FunctionList -> NodeList Int
numberFrom n xs = zipWith (\x (F f) -> N x f) [n..] xs

-- | Create edges that carry no filter function
toEL :: [SimpleEdge n l] -> EdgeList n l
-- the 'Nothing' needs a concrete type, but this is not relevant 
toEL = map (\(x, y, z) -> E x y z (Nothing :: Maybe (Int -> Int)))

-- | Drops the filter function from the edges
fromEL :: EdgeList n l -> [SimpleEdge n l]
fromEL = map (\(E a b c _) -> (a, b, c))

-- | A list of 'Nothing'. 
nothings :: [Maybe (Int -> Int)]
nothings = repeat Nothing


---------------------------------------------------------------------
-- getting multiple inputs into a single list

-- | The 'Lister' data type represents a placeable function for which those
-- arguments which are of type list may either be received as a stream (as 
-- in the ordinary case) or may be received as single elements via different 
-- channels.
data Lister f = Lister [Int] f

-- | Create a 'Lister' from a function. The second parameter gives the information 
-- how each parameter of the function should be received. A 0 indicates ordinary 
-- behaviour, a positive value gives the amount of elements in the list parameter. 
-- A positive value for non-list parameters is ignored. Trailing zeros may be omitted.
lister :: (IsFunctionType f flag, Placeable' flag f a g r p) =>  f -> [Int] -> Lister f
lister f ns = Lister ns f

-- | The instance for the 'Lister' type
instance (IsFunctionType ftype flag,
          Placeable' flag ftype argtype remtype restype plisttype) => 
         Placeable (Lister ftype) argtype remtype restype plisttype where
        createChan (Lister ns f) = createChan' (undefined::flag) f
        apply (Lister ns f) = apply' (undefined::flag) f
        castOutputChan (Lister ns f) = castOutputChan' (undefined::flag) f
        isConst (Lister ns f) = isConst' (undefined::flag) f
        applyAll (Lister ns f) = applyAll' (undefined::flag) f
        generateChannels (Lister ns f) = generateListerChannels' (undefined::flag) f ns

---------------------------------------------------------------------
-- graph --> process network

-- | A dummy data type.
-- A ('ChanName' a) will be cast into ('ChanName' 'Dummy') for transport 
data Dummy
instance NFData Dummy
instance Trans Dummy


-- a type for communication when building the network 
data BuildMsg = ChanMsg Int Int Int (ChanName Dummy)

instance NFData BuildMsg where
    rnf (ChanMsg from to edgeId chan) = rnf from `seq` rnf to `seq` rnf edgeId `seq` rnf chan
instance Trans BuildMsg

instance Show BuildMsg where
    show (ChanMsg from to edgeId name) = "ChanMsg "++(show from)++" "++show to++" "++show edgeId


-- | The abstraction of a process network that computes a result of type a on its main node.
data ProcessNetwork a = Net ([[BuildMsg]] -> (a, [[BuildMsg]])) [Process ((), [BuildMsg]) [BuildMsg]]


-- | Builds a 'ProcessNetwork' from a given graph-specification. 
-- The graph is specified as a list of nodes and a list of edges.
build :: forall f a g r p n e. (Placeable f a g r p, Ord e, Eq n) => (n, f) -> NodeList n -> [Edge n e] -> ProcessNetwork r
build (label, func) nodes edges = Net mainFunction processList
    where
        mainFunction :: [[BuildMsg]] -> (r, [[BuildMsg]])
        mainFunction mss = (result, [filter (\(ChanMsg _ t _ _) -> t == x) ms | x <- [1..length nodes]])
            where
                ms = handleMsgs 0 label (length edges) func result ((toChanMsgs 0 label myInChannels) ++ merge mss)
                --labelList = map (\(N l _) -> l) nodes
                (myInChannels, myParamList) = generateChannels func
                result = applyAll func myParamList

        processList :: [Process ((), [BuildMsg]) [BuildMsg]]
        processList = map (process.procFunc) (zip [1..] nodes)

        edgeTable = zip [0..] edges
        
        --getNodeId :: n -> Int
        --getNodeId n = fromJust.elemIndex n nodeList
        allLabels = label : map (\(N l _) -> l) nodes

        --                   (edgeId, chanReceiverId)
        paramGivers :: n -> [(Int, Int)]
        paramGivers n = (proj.sort.filt) edgeTable
            where
                proj = map (\(i, E x _ _ _) -> (i,  fromJust $ elemIndex x allLabels))
                sort = sortBy (\(_, E _ _ x _) (_, E _ _ y _) -> compare x y)
                filt = filter (\(_, E _ x _ _) -> x == n)
        
        
        outEdgeTable :: n -> [(Int, Edge n e)]
        outEdgeTable n = filter (\(_, E x _ _ _) -> x == n) edgeTable

        toChanMsgs :: Int -> n -> [ChanName Dummy] -> [BuildMsg]
        toChanMsgs nId n cs = {-trace ("ChanMsg construction paramGivers="++show (paramGivers n)) $-} zipWith (uncurry (flip $ ChanMsg nId)) (paramGivers n) cs

        handleMsgs :: (Placeable f' a' g' r' p') => Int -> n -> Int -> f' -> r' -> [BuildMsg] -> [BuildMsg]
        handleMsgs nId n k df result ms = handleMsgs' k ms
            where
                handleMsgs' :: Int -> [BuildMsg] -> [BuildMsg]
                handleMsgs' 0 nil = if n == label then ({-trace (show n++" delaying finish until result") () `seq`-} rnf nil `seq` {-trace (show n++" finished")-} []) else []
                handleMsgs' k ((m@(ChanMsg from to edgeId chan)):xs) 
                    | to == nId = {-trace (show n++" received chan from "++show from++" edgeId="++show edgeId) $-}
                                  case (lookup edgeId (outEdgeTable n)) of
                                      (Just (E _ _ _ Nothing)) -> {-trace (show n ++ " filling all") $ -}
                                                                  parfill c result (handleMsgs' (k-1) xs)
                                      (Just (E _ _ _ (Just s))) -> {-trace (show n++" filling selection") $ -}
                                                                   parfill (castOutputChan s chan) (apply s (unsafeCoerce result)) (handleMsgs' (k-1) xs)
                                      Nothing -> {-trace (show n++" filling all") $ -}
                                                 parfill c result (handleMsgs' (k-1) xs)
                    | otherwise = {-trace (show n++" passing "++show m) $ -}
                                  m:(handleMsgs' (k-1) xs)
                        where
                                c = (castOutputChan df chan)


        procFunc :: (Int, Node n) -> ((), [BuildMsg]) -> [BuildMsg]
        procFunc (nId, N n df) ((), incomingMsgs) = {-trace (show n++" start "++show (toChanMsgs n myInChannels)) $-}
                                                            (toChanMsgs nId n myInChannels) ++ (handleMsgs nId n (length (outEdgeTable n)) df result incomingMsgs)
            where
                (myInChannels, myParamList) = generateChannels df

                result = {-trace (show n++" result") $-} applyAll df myParamList

-- | Instantiates the given 'ProcessNetwork' and computes its result.
start :: (Trans a) => ProcessNetwork a -> a
start (Net f ps) = rnf xs `seq` result
    where
        (result, toNodes) = f xs

        xs :: [[BuildMsg]]
        xs = runPA $ zipWithM instantiate ps input

        input :: [((), [BuildMsg])]
        input = lazy2Zip (replicate (length ps) ()) toNodes


lazy2Zip (x:xs) ~(y:ys) = (x,y): lazy2Zip xs ys
lazy2Zip []     _     = []

---------------------------------------------------------------------
-- checks

-- | Poor man's substitute for typechecking: Checks the validity of a graph structure by checking
-- that each function that is placed on a node gets exactly as many arguments via incoming edges as it needs.
-- The result is a list of error messages that identify where this property is violated. 
-- An empty list as result indicates validity.
checkTopology :: (Placeable f a g r p, Show e) => (Int, f) -> NodeList Int -> EdgeList Int e -> [String]
checkTopology (l, f) ns edges = (compareCounts arities incoming) ++ (reportInvalidEdges edges)
    where
        nodes = (N l f):ns

        --compareCounts checks whether each function gets the expected number of arguments
        compareCounts :: [(Int, Int)] -> [(Int, Int)] -> [String]
        compareCounts [] _ = []
        compareCounts ((l1, 0):fs) [] = compareCounts fs []
        compareCounts ((l1, c1):fs) [] = ("function "++show l1++" expects "++show c1++" arguments but gets 0"):compareCounts fs []
        compareCounts ((l1, c1):fs) ((l2, c2):es) 
            | l1 < l2 = ("function "++show l1++" expects "++show c1++" arguments but gets 0"):compareCounts fs ((l2, c2):es)
            | l1 > l2 = compareCounts ((l1, c1):fs) es
            | l1 == l2 && c1 /= c2 = ("function "++show l1++" expects "++show c1++" arguments but gets "++show c2):compareCounts fs es
            | l1 == l2 && c1 == c2 = compareCounts fs es

        arities :: [(Int, Int)]
        arities = sortBy (\x y -> compare (fst x) (fst y)) $ map (\(N l f) -> (l, (length.fst.generateChannels) f)) nodes
        incoming :: [(Int, Int)]
        incoming = (map count.grp.sort) edges

        sort = sortBy (\(E _ t1 _ _) (E _ t2 _ _) -> compare t1 t2)
        grp = groupBy (\(E _ t1 _ _) (E _ t2 _ _) -> t1 == t2)
        count ((E _ t _ _):xs) = (t, (1 + length xs))

        reportInvalidEdges [] = []
        reportInvalidEdges (e@(E f t l _):es) 
            | f `elem` nodeLabels && t `elem` nodeLabels = reportInvalidEdges es
            | f `elem` nodeLabels = ("edge "++show e++" references non-existing node "++show t):reportInvalidEdges es
            | t `elem` nodeLabels = ("edge "++show e++" references non-existing node "++show f):reportInvalidEdges es
            | otherwise = ("edge "++show e++" references non-existing nodes "++show f++", "++show t):reportInvalidEdges es

        nodeLabels = map (\(N l _) -> l) nodes

-- Convenience predicate for checking the validity of a graph structure. It applies 'checkTopology' and
-- compares its result to the empty list.
isTopologyOK :: (Placeable f a g r p, Show e) => (Int, f) -> NodeList Int -> [Edge Int e] -> Bool
isTopologyOK c ns es = (==[]) (checkTopology c ns es)

